
import os
import yaml

from elasticsearch_kibana_cli import __title__ as NAME
from elasticsearch_kibana_cli.utils.logger import Logger
from elasticsearch_kibana_cli.exceptions.ElasticsearchKibanaCLIException import ElasticsearchKibanaCLIException


logger = Logger(name=NAME).logging


class ElasticsearchKibanaCLIConfig:

    config = None
    config_root = NAME
    config_filename = None

    def __init__(self, config_filename=None):

        self.config_filename = os.path.expanduser(config_filename)
        logger.debug('loading config_filename: {}'.format(self.config_filename))

        if self.config_filename is None or not os.path.isfile(self.config_filename):
            raise ElasticsearchKibanaCLIException('Unable to locate config_filename', self.config_filename)

        self.config = self.__load_config(self.config_filename)

    def __load_config(self, config_filename):
        loaded_config = {}

        with open(config_filename, 'r') as f:
            try:
                loaded_config = yaml.safe_load(f.read().replace('@timestamp', '___timestamp'))
            except yaml.YAMLError as e:
                raise ElasticsearchKibanaCLIException(e)

        loaded_config = self.__replace_env_values(loaded_config)

        if type(loaded_config) is not dict:
            raise ElasticsearchKibanaCLIException('Configuration loaded is incorrect data type', self.config_root)

        if type(loaded_config) is not dict or self.config_root not in loaded_config.keys():
            raise ElasticsearchKibanaCLIException('Unable to locate config root', self.config_root)

        return loaded_config[self.config_root]

    def __replace_env_values(self, input):
        if input is None:
            return input
        elif type(input) in (int, bool):
            return input
        elif type(input) is str:
            if input.lower()[0:4] == 'env:':
                env_name = input.replace('env:', '')
                logger.debug('Config element set via env value {}'.format(env_name))
                value = os.getenv(env_name, None)
                if value is None or len(value) < 1:
                    raise ElasticsearchKibanaCLIException('Config requested env value not set', env_name)
                return value
            return input
        elif type(input) is list:
            r = []
            for item in input:
                r.append(self.__replace_env_values(item))
            return r
        elif type(input) is dict:
            r = {}
            for item_k, item_v in input.items():
                r[item_k] = self.__replace_env_values(item_v)
            return r
        else:
            raise ElasticsearchKibanaCLIException('Unsupported type in replace_env_values()', input)




import os
import click

from elasticsearch_kibana_cli import __title__ as NAME
from elasticsearch_kibana_cli import __version__ as VERSION
from elasticsearch_kibana_cli import __env_config_filename__ as ENV_CONFIG_FILENAME
from elasticsearch_kibana_cli import __search_split_count_default__ as SEARCH_SPLIT_COUNT_DEFAULT

from elasticsearch_kibana_cli.utils.logger import Logger
from elasticsearch_kibana_cli.utils.output import output_handler
from elasticsearch_kibana_cli.ElasticsearchKibanaInterface import ElasticsearchKibanaInterface


config_file_default = '~/.eskbcli'
elasticsearch_kibana_interface = None


@click.group()
@click.option('-c', '--config', help='Config file location; overrides {} environment var and the default {} value.'
              .format(ENV_CONFIG_FILENAME, config_file_default))
@click.option('-v', '--verbose', is_flag=True, help='Verbose logging messages (debug level).')
@click.option('-q', '--quiet', is_flag=True, help='Quiet mode, takes priority over --verbose')
@click.version_option(VERSION)
def eskbcli_interface(config, verbose, quiet):
    """
    ElasticSearch Kibana CLI (`eskbcli`) provides a shell interface to query
    an ElasticSearch backend via the Kibana frontend which is useful in
    situations where the ElasticSearch backend is not otherwise accessible.

    ElasticSearch Kibana CLI makes it possible to copy-paste query expressions
    directly from the Kibana user-interface and then easily access very large
    sets of result data.  This makes the `eskbcli` useful in SecOps situations
    where the ability to rapidly move from a Kibana query to raw data is
    valued.

    Configuration options are available to adjust http-headers so-as-to enable
    access to Kibana in situations that require complex user-authentication
    such as when Kibana exists behind an OAuth reverse proxy or other session-
    based authentication arrangement.

    Documentation available https://elasticsearch-kibana-cli.readthedocs.io
    """

    if quiet:
        logger = Logger(name=NAME).setup(level='CRITICAL')
    elif verbose:
        logger = Logger(name=NAME).setup(level='DEBUG')
    else:
        logger = Logger(name=NAME).setup(level='INFO')

    if config is not None:
        config_filename = config
        logger.debug('config_filename taken from cli arg: {}'.format(config_filename))
    else:
        if os.getenv(ENV_CONFIG_FILENAME):
            config_filename = os.getenv(ENV_CONFIG_FILENAME)
            logger.debug('config_filename taken from env value: {}'.format(config_filename))
        else:
            config_filename = config_file_default
            logger.debug('config_filename using default: {}'.format(config_filename))

    global elasticsearch_kibana_interface
    elasticsearch_kibana_interface = ElasticsearchKibanaInterface(config_filename=config_filename)


@eskbcli_interface.command('list')
@click.option('-o', '--out', help='Filename to write data output.', required=False)
def list_searches(**kwargs):
    """
    List the available eskbcli search names.
    """
    output_handler(
        data=elasticsearch_kibana_interface.list_searches(),
        filename=kwargs['out'], compact=False if kwargs['out'] is None else True
    )


@eskbcli_interface.command('show')
@click.option('-o', '--out', help='Filename to write data output.', required=False)
@click.argument('search_name', required=True)
def show_search(**kwargs):
    """
    Show the named eskbcli search configuration.
    """
    output_handler(
        data=elasticsearch_kibana_interface.show_search(name=kwargs['search_name']),
        filename=kwargs['out'], compact=False if kwargs['out'] is None else True
    )


@eskbcli_interface.command('search')
@click.option('-o', '--out', help='Filename to write data output.', required=False)
@click.option('-s', '--splits', help='Number of splits to break the search into (default:{})'
              .format(SEARCH_SPLIT_COUNT_DEFAULT), required=False)
@click.option('-np', '--no-ping', help='Do not ping the Kibana endpoint before using the connection (default: False)',
              is_flag=True, required=False, default=False)
@click.argument('search_name', required=True)
def perform_search(**kwargs):
    """
    Execute the named search configuration.
    """
    output_handler(
        data=elasticsearch_kibana_interface.perform_search(
            name=kwargs['search_name'],
            split_count=kwargs['splits'],
            ping_connection=False if kwargs['no_ping'] is True else True
        ),
        filename=kwargs['out'],
        compact=False if kwargs['out'] is None else True
    )




import os
import logging
import requests
from flask import Flask, Response, request


class ElasticsearchKibanaCLIInternalProxy:

    debug = False
    listen_port = 59200
    listen_proto = 'http'
    listen_address = '127.0.0.1'
    header_overrides = {}
    request_methods = ['GET', 'POST']

    client_connect_address = None

    def __init__(self, config):

        for config_k, config_v in config.items():
            if config_k.lower() == 'header_overrides':
                self.header_overrides = config_v
            elif config_k.lower() == 'listen':
                if 'address' in config_v.keys():
                    self.listen_address = config_v['address']
                if 'port' in config_v.keys():
                    self.listen_port = config_v['port']
                if 'proto' in config_v.keys():
                    self.listen_proto = config_v['proto']
            elif config_k.lower() == 'debug' and str(config_v).lower() in ['yes', 'true', 'on', '1']:
                self.debug = True

        self.client_connect_address = '{}://{}:{}'.format(self.listen_proto, self.listen_address, self.listen_port)

    def start(self, base_uri):

        app = Flask('ElasticsearchKibanaCLIInternalProxy')

        if not self.debug:
            os.environ['WERKZEUG_RUN_MAIN'] = 'true'
            logger = logging.getLogger('werkzeug')
            logger.disabled = True

        @app.route('/<path:path>', methods=self.request_methods)
        @app.route('/', defaults={'path': None}, methods=self.request_methods)
        def catchall(path):

            method = request.method
            path = request.full_path
            data = request.get_data(cache=False, as_text=False, parse_form_data=False)

            headers = {}
            headers_replaced = []
            for header in str(request.headers).split('\r\n'):
                if ': ' in header:
                    header_k, header_v = header.split(': ', 1)
                    if header_k.lower() in self.header_overrides.keys():
                        headers[header_k.lower()] = self.header_overrides[header_k.lower()]
                        headers_replaced.append(header_k.lower())
                    else:
                        headers[header_k.lower()] = header_v

            for header_k in self.header_overrides.keys():
                if header_k.lower() not in headers_replaced:
                    headers[header_k.lower()] = self.header_overrides[header_k]

            if 'cookie' in headers.keys():
                headers['cookie'] = headers['cookie'].replace('cookie: ', '').replace('Cookie: ', '').strip()

            if base_uri[-1:] == '/' and path[:1] == '/':
                url = '{}{}'.format(base_uri, path[1:])
            elif base_uri[-1:] != '/' and path[:1] != '/':
                url = '{}/{}'.format(base_uri, path)
            else:
                url = '{}{}'.format(base_uri, path)

            # print([url, headers, data])
            r = requests.request(method, url=url, headers=headers, data=data, stream=True)
            return Response(r.content, status=r.status_code)

        app.run(host=self.listen_address, port=self.listen_port, threaded=True, debug=False)




import sys
import json

from elasticsearch_kibana_cli import __title__ as NAME
from elasticsearch_kibana_cli.utils.logger import Logger


logger = Logger(NAME).logging


def output_handler(data, filename=None, compact=False):

    if compact:
        output = json.dumps(data, indent=None, separators=(',', ':'))
    else:
        output = json.dumps(data, indent='  ', separators=(', ', ': '))

    if filename is None or filename.upper() == 'STDOUT':
        print(output)
    elif filename.upper() == 'STDERR':
        print(output, file=sys.stderr)
    else:
        logger.info('Saving output to filename {}'.format(filename))
        with open(filename, 'w') as f:
            f.write(output)




import json
import time
import maya
import copy
import requests
import dpath.util
from collections import OrderedDict
from elasticsearch_dsl import Q

from elasticsearch_kibana_cli import __title__ as NAME
from elasticsearch_kibana_cli import __version__ as VERSION
from elasticsearch_kibana_cli import __search_split_bucket_limit__ as SEARCH_SPLIT_LIMIT
from elasticsearch_kibana_cli import __search_default_timeout_seconds__ as SEARCH_DEFAULT_TIMEOUT_SECONDS


from elasticsearch_kibana_cli.exceptions.ElasticsearchKibanaCLIException import ElasticsearchKibanaCLIException
from elasticsearch_kibana_cli.utils.logger import Logger


logger = Logger(name=NAME).logging


class ElasticsearchKibanaCLISearch:

    connection = None
    user_agent = '{}/{}'.format(NAME, VERSION)

    def __init__(self, connection):
        self.connection = connection

    def msearch(self, index, search, aggs=None, size=SEARCH_SPLIT_LIMIT, source=None, splits=1, range_keyword='range'):

        url = '{}/elasticsearch/_msearch'.format(self.connection.client_connect_address)

        payload_data = ''
        for _ in range(0, splits):
            payload_data = '{}{}\n{}\n'.format(
                payload_data,
                self.__payload_header(copy.copy(index)),
                self.__payload_body(
                    query_params=copy.copy(search),
                    aggs_params=copy.copy(aggs),
                    size=copy.copy(size),
                    source=copy.copy(source)
                )
            )

        if splits > 1:
            logger.debug('Splitting search into {} requests based on "{}" keyword'.format(str(splits), range_keyword))
            payload_data = self.__payload_range_splitter(payload_data, range_keyword=range_keyword)
            logger.info('Search split into {} requests based on "{}" keyword'.format(str(splits), range_keyword))

        request_headers = {
            'content-type': 'application/x-ndjson',
            'kbn-version': str(self.connection.kbn_version),
            'user-agent': self.user_agent
        }

        r = requests.post(
            url,
            data=payload_data,
            headers=request_headers
        )

        if r.status_code >= 300:
            logger.error('Unexpected http-response code {}'.format(r.status_code))
            if len(r.text) > 1:
                logger.error(str(r.text.replace("\r", "").replace("\n", ""))[:128])
            return []

        try:
            response_data = r.json()
        except json.decoder.JSONDecodeError as e:
            logger.error(e)
            if len(r.text) > 1:
                logger.error(str(r.text.replace("\r", "").replace("\n", ""))[:128])
            return []

        return_list = []
        hit_min = 9999999999
        hit_max = 0
        hit_total = 0

        for hit_index in range(0, splits):
            value = None
            try:
                value = response_data['responses'][hit_index]['hits']['total']  # responses/0/hits/total
                return_list.extend(response_data['responses'][hit_index]['hits']['hits'])
            except KeyError:
                pass
            except IndexError:
                pass

            if value is None:
                value = 0
            elif value > SEARCH_SPLIT_LIMIT:
                logger.warning('Search split {} has {} hit-results which exceeds the {} limit, '
                               'results truncated!'.format(hit_index, value, SEARCH_SPLIT_LIMIT))

            if value < hit_min:
                hit_min = value
            if value > hit_max:
                hit_max = value

            hit_total = hit_total + value

        logger.info('results-available={available_count}; results-returned={returned_count}; '
                    'min/max/avg result-count-per-split={min_split_count}/{max_split_count}/{avg_split_count}; '
                    'msearch-splits={split_count}'.format(
                        available_count=hit_total,
                        returned_count=len(return_list),
                        min_split_count=hit_min,
                        max_split_count=hit_max,
                        avg_split_count=int(hit_total / splits),
                        split_count=splits,
                    ))

        if self.connection.internal_proxy:
            time.sleep(0.1)  # allows internal_proxy threads to close-out

        return return_list

    def __payload_header(self, index):
        return json.dumps({
            'index': index,
            'ignore_unavailable': True,
            'preference': int(round(time.time() * 1000))
        })

    def __payload_body(self, query_params, aggs_params, size=SEARCH_SPLIT_LIMIT, source=None):

        if size > SEARCH_SPLIT_LIMIT or size < 1:
            raise ElasticsearchKibanaCLIException('Payload size is out-of-bounds in __payload_body()', size)

        for param_name in ['must', 'must_not', 'should', 'should_not', 'filter']:
            if param_name in query_params:
                query_params[param_name] = self.__parse_query_param(query_params[param_name])
            else:
                query_params[param_name] = []

        aggs = {}  # for another day

        query = Q(
            'bool',
            must=query_params['must'],
            must_not=query_params['must_not'],
            should=query_params['should'],
            should_not=query_params['should_not'],
            minimum_should_match=query_params['minimum_should_match'] if 'minimum_should_match' in query_params else (
                1 if len(query_params['should']) > 0 else None
            ),
            filter=query_params['filter']
        )

        payload_values = {
            'source': json.dumps(source).replace('___timestamp', '@timestamp') if source is not None else '[ ]',
            'size': size,
            'aggs': json.dumps(aggs),
            'query': json.dumps(query.to_dict()),
            'timeout': '"' + (str(query_params['timeout']) if 'timeout' in query_params else '{}s'.
                              format(SEARCH_DEFAULT_TIMEOUT_SECONDS)) + '"'
        }

        payload_json = """
            {
              "version": true,
              "sort": [ ],
              "stored_fields": [ ],
              "script_fields": { },
              "docvalue_fields": [ ],
              "highlight": { },
              "_source": __SOURCE__,
              "size": __SIZE__,
              "aggs": __AGGS__,
              "query": __QUERY__,
              "timeout": __TIMEOUT__
            }
        """

        for payload_k, payload_v in payload_values.items():
            replace_token = '__{}__'.format(payload_k.upper())
            payload_json = payload_json.replace(replace_token, str(payload_v))

        return json.dumps(json.loads(payload_json))

    def __parse_query_param(self, query_param):
        query = []
        for param_query_type, param_queries in query_param.items():
            for param_query_item in param_queries:
                if param_query_item == '___timestamp':
                    timestamp = self.__parse_query_timestamp(param_queries[param_query_item])
                    query.append(Q(param_query_type, **{'@timestamp': timestamp}))
                elif type(param_query_item) is dict:
                    for param_query_item_k, param_query_item_v in param_query_item.items():
                        query.append(Q(param_query_type, **{param_query_item_k: param_query_item_v}))
                else:
                    raise ElasticsearchKibanaCLIException('Unsupported type in __parse_query_param()', param_query_item)
        return query

    def __parse_query_timestamp(self, element):
        if type(element) in [int, str, bool] or element is None:
            return element
        elif type(element) is list:
            r = []
            for item in element:
                r.append(self.__parse_query_timestamp(item))
            return r
        elif type(element) is dict:
            r = {}
            for key, value in element.items():
                if key in ['gt', 'gte', 'lt', 'lte', 'eq']:
                    r[key] = int(maya.when(value).datetime().timestamp() * 1000)
                else:
                    r[key] = self.__parse_query_timestamp(value)
            r['format'] = 'epoch_millis'
            return r
        else:
            raise ElasticsearchKibanaCLIException('Unsupported type in __parse_query_timestamp()', element)

    def __payload_range_splitter(self, ndjson_payload, range_keyword='range'):

        return_payloads = OrderedDict()
        for index, json_payload in enumerate(ndjson_payload.split('\n')):
            if len(json_payload) > 0:
                return_payloads[index] = json.loads(json_payload)

        payload_ranges = OrderedDict()
        for payload_k, payload_v in return_payloads.items():
            for payload_tupple in dpath.util.search(payload_v, '**/{}'.format(range_keyword), yielded=True):
                payload_ranges[payload_k] = payload_tupple

        # check the range paths are the same
        payload_range_path = None
        payload_range_path_key = None
        for payload_range_k, payload_range_value in payload_ranges.items():
            if payload_range_path is None:
                payload_range_path, payload_range_path_value = payload_range_value
                payload_range_path_key = next(iter(payload_range_path_value.keys()))
            else:
                payload_range_path_next, payload_range_path_next_value = payload_range_value
                payload_range_path_next_key = next(iter(payload_range_path_next_value.keys()))
                if payload_range_path_next != payload_range_path:
                    raise ElasticsearchKibanaCLIException(
                        'Ranges with different paths in __payload_range_splitter()'
                    )
                if payload_range_path_next_key != payload_range_path_key:
                    raise ElasticsearchKibanaCLIException(
                        'Ranges with different internal keys in __payload_range_splitter()'
                    )

        # determine the min max values within the discovered range definitions
        min_value = max_value = None
        limit_keys = []
        for range_key in payload_ranges.keys():
            range_values = dpath.util.get(
                return_payloads,
                '{}/{}/{}'.format(range_key, payload_range_path, payload_range_path_key)
            )
            for range_value_key, range_value_value in range_values.items():
                if range_value_key.lower() in ['gt', 'gte', 'lt', 'lte']:
                    if min_value is None:
                        min_value = range_value_value
                    elif range_value_value < min_value:
                        min_value = range_value_value
                    if max_value is None:
                        max_value = range_value_value
                    elif range_value_value > max_value:
                        max_value = range_value_value
                    if range_value_key not in limit_keys:
                        limit_keys.append(range_value_key)

        def split_min_max_delta(min_max_delta, num_of_parts):
            part_duration = int(min_max_delta / num_of_parts)
            parts = []
            marker = 0
            for _ in range(num_of_parts):
                part = [marker, marker + part_duration]
                marker += part_duration
                parts.append(part)
            parts[len(parts) - 1][1] = min_max_delta  # rewrite final value so it correctly lines up
            return parts

        splits = split_min_max_delta(int(max_value - min_value), len(payload_ranges.keys()))

        splits_index = 0
        for range_key in payload_ranges.keys():
            for limit_key in limit_keys:

                if splits_index == 0 and 'gt' in limit_key.lower():
                    dpath.util.set(
                        return_payloads,
                        '{}/{}/{}/{}'.format(range_key, payload_range_path, payload_range_path_key, limit_key),
                        (min_value + splits[splits_index][0])
                    )

                elif 'gt' in limit_key.lower():
                    dpath.util.delete(
                        return_payloads,
                        '{}/{}/{}/{}'.format(range_key, payload_range_path, payload_range_path_key, limit_key)
                    )
                    dpath.util.new(
                        return_payloads,
                        '{}/{}/{}/gte'.format(range_key, payload_range_path, payload_range_path_key),
                        (min_value + splits[splits_index][0])
                    )

                elif splits_index == len(splits) - 1 and 'lt' in limit_key.lower():
                    dpath.util.set(
                        return_payloads,
                        '{}/{}/{}/{}'.format(range_key, payload_range_path, payload_range_path_key, limit_key),
                        (min_value + splits[splits_index][1])
                    )

                elif 'lt' in limit_key.lower():
                    dpath.util.delete(
                        return_payloads,
                        '{}/{}/{}/{}'.format(range_key, payload_range_path, payload_range_path_key, limit_key)
                    )
                    dpath.util.new(
                        return_payloads,
                        '{}/{}/{}/lt'.format(range_key, payload_range_path, payload_range_path_key),
                        (min_value + splits[splits_index][1])
                    )

                else:
                    raise ElasticsearchKibanaCLIException(
                        'Unsupported limit key in __payload_range_splitter()', limit_key
                    )

            splits_index += 1

        # re-cast back into ndjson_payload format
        ndjson_return_payload = ''
        for return_payloads_k, return_payloads_v in return_payloads.items():
            ndjson_return_payload = '{}{}\n'.format(ndjson_return_payload, json.dumps(return_payloads_v))

        # logger.debug(ndjson_return_payload)
        return ndjson_return_payload




import time
import logging


class LoggerException(Exception):
    pass


class Logger:

    name = None
    logging = None
    default_level = 'info'

    def __init__(self, name):
        self.name = name
        self.logging = logging.getLogger(self.name)

    def setup(self, level=None, fmt='%(asctime)s - %(levelname)s - %(message)s', datefmt='%Y%m%dZ%H%M%S'):

        logger_init = logging.getLogger(self.name)
        stream_handler = logging.StreamHandler()

        if level is not None:
            log_level = level.upper()
        else:
            log_level = self.default_level.upper()

        if log_level == 'CRITICAL':
            logger_init.setLevel(logging.CRITICAL)
            stream_handler.setLevel(logging.CRITICAL)
        elif log_level == 'ERROR':
            logger_init.setLevel(logging.ERROR)
            stream_handler.setLevel(logging.ERROR)
        elif log_level in ('WARNING', 'WARN'):
            logger_init.setLevel(logging.WARNING)
            stream_handler.setLevel(logging.WARNING)
        elif log_level == 'INFO':
            logger_init.setLevel(logging.INFO)
            stream_handler.setLevel(logging.INFO)
        elif log_level == 'DEBUG':
            logger_init.setLevel(logging.DEBUG)
            stream_handler.setLevel(logging.DEBUG)
        else:
            raise LoggerException('unknown logger level value', log_level)

        logging.Formatter.converter = time.gmtime
        stream_handler.setFormatter(LoggerColorFormatter(fmt=fmt, datefmt=datefmt))
        logger_init.addHandler(stream_handler)

        self.logging = logger_init
        return self.logging


class LoggerColorFormatter(logging.Formatter):

    color_line = '\x1b[90m'  # grey
    color_reset = '\x1b[0m'  # reset

    def __init__(self, **kwargs):
        if 'fmt' in kwargs:
            kwargs['fmt'] = '{}{}{}'.format(self.color_line, kwargs['fmt'], self.color_reset)
        logging.Formatter.__init__(self, **kwargs)

    def format(self, record):

        levelname = record.levelname.upper()

        if levelname == 'CRITICAL':
            color_code = '\x1b[41m'  # white-on-red
        elif levelname == 'ERROR':
            color_code = '\x1b[31m'  # red
        elif levelname in ('WARNING', 'WARN'):
            color_code = '\x1b[33m'  # yellow
        elif levelname == 'INFO':
            color_code = '\x1b[36m'  # cyan
        elif levelname == 'DEBUG':
            color_code = '\x1b[37m'  # white
        else:
            color_code = '\x1b[90m'  # grey

        record.levelname = '{}{}{}{}'.format(color_code, levelname, self.color_reset, self.color_line)
        return logging.Formatter.format(self, record)
